package ru.alphatech.icfpc.game;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import ru.alphatech.icfpc.cards.Dbl;
import ru.alphatech.icfpc.cards.ICard;
import ru.alphatech.icfpc.cards.K;
import ru.alphatech.icfpc.cards.Put;
import ru.alphatech.icfpc.cards.S;
import ru.alphatech.icfpc.cards.Succ;
import ru.alphatech.icfpc.cards.Zero;
import ru.alphatech.icfpc.game.Batch.OperationType;

/**
 * 
 * TODO 
 * 
 * @author akhmetovlz
 *
 */
public abstract class AbstractPolicy {

	protected ArrayList<Batch> batches = new ArrayList<Batch>();
	
	public AbstractPolicy(){
	}	
	
	public void setRegister(int registerIndex, int num ){
		if( registerIndex < 4) {
			Slot registerSlot = Run.getRegisterByIndex(registerIndex);
			batches.addAll( numberIntoSlotField( registerSlot, num ) );
		}
	}
	
	protected ArrayList<Batch> numberIntoSlotField(Slot targetSlot, int number)
	{	
		ArrayList<Batch> retArray = new ArrayList<Batch>();
		
		if(number == 0){
			retArray.add(new Batch(OperationType.CARD_TO_SLOT, new Put(), targetSlot));
			retArray.add(new Batch(OperationType.SLOT_TO_CARD, new Zero(), targetSlot));
		}else{
			retArray.addAll(smartAlg(0, number, targetSlot));
		}
		return retArray;
	}
	
	
	protected ArrayList<Batch> stupidAlg(int begin, int end, Slot targetSlot){
		ArrayList<Batch> retArray = new ArrayList<Batch>();
		while( begin*2 < end){
			retArray.add(new Batch(OperationType.CARD_TO_SLOT, new Dbl(), targetSlot));
			begin *= 2;
		}
		while( begin != end){
			retArray.add(new Batch(OperationType.CARD_TO_SLOT, new Succ(), targetSlot));
			begin ++;
		}
		return retArray;
	}
	
	protected static List<Batch> smartAlg(int begin, int end, Slot targetSlot){
		ArrayList<Batch> retArray = new ArrayList<Batch>();
		
		if(begin == 0){
			retArray.add(new Batch(OperationType.CARD_TO_SLOT, new Put(), targetSlot));
			retArray.add(new Batch(OperationType.SLOT_TO_CARD, new Zero(), targetSlot));
			retArray.add(new Batch(OperationType.CARD_TO_SLOT, new Succ(), targetSlot));
			
			if(end == 0)
				return retArray;
		}
		
		List<Batch> list1 = buildNumberOperations(0, end, targetSlot);
		List<Batch> list2 = buildNumberOperations(begin, end, targetSlot);
		if (list1.size() <= list2.size())
		{
			Collections.reverse(list1);
			retArray.addAll(list1);
		}
		else
		{
			Collections.reverse(list2);
			retArray.addAll(list2);
		}
		return retArray;
	}
	
	protected static List<Batch> buildNumberOperations(int begin, int end, Slot targetSlot){
		List<Batch> list = new ArrayList<Batch>();
		
		while(end >= 2){
			int d = end % 2;
			if (begin > 0 && 2*begin > end)
			{
				for (int i = 0; i < end - begin; i++)
				{
					list.add(new Batch(OperationType.CARD_TO_SLOT, new Succ(), targetSlot));
				}
				break;
			}
			if( d == 0 ) {
				list.add(new Batch(OperationType.CARD_TO_SLOT, new Dbl() , targetSlot));
				end = end / 2;
			} else {
				list.add(new Batch(OperationType.CARD_TO_SLOT, new Succ(), targetSlot));
				list.add(new Batch(OperationType.CARD_TO_SLOT, new Dbl() , targetSlot));
				end = (end-1) / 2;
			}
		}
		return list;
	}
	
	/**
	 * Корректировка позиций регистров
	 * В самом лучшем случаи регистры это слоты 0 1 2 3 
	 * Но могут быть изменены, этот метод корректирует ссылки на регистры в полтиках
	 * @param registerIndex - индекс регитсра (не слота!)
	 * @param processingSlot - слот где происходит рассчёт
	 * @return
	 */
	protected ArrayList<Batch> correctRegister(int registerIndex, Slot processingSlot){
		ArrayList<Batch> retArray = new ArrayList<Batch>();
		for(int i=0; i<Run.getRegisterSlotNum(registerIndex); ++i) 
		{
			retArray.add(new Batch(OperationType.CARD_TO_SLOT, new K(), processingSlot));
			retArray.add(new Batch(OperationType.CARD_TO_SLOT, new S(), processingSlot));
			
			retArray.add(new Batch(OperationType.SLOT_TO_CARD, new Succ(), processingSlot));
		}
		return retArray;
	}
	
	public abstract ArrayList<Batch> getBatches();
	
	public static void main(String[] args)
	{
		Slot slot = new Slot(1);
		List<Batch> list = smartAlg(0, 8, slot);
		System.out.println(list.toArray().toString());
	}
}
